import datetime
import json
from typing import Any

from sqlalchemy import Table
from sqlalchemy.ext.declarative import DeclarativeMeta

from . import SQLAIntrospector


class SqlaDeclarativeEncoder(json.JSONEncoder):
    """
    A custom JSON encoder for serializing SQLAlchemy
    declarative base instances.

    Supports ISO8601 compliant datetime encoding.
    """
    datetypes = (
        datetime.date,
        datetime.datetime,
        datetime.timedelta
    )

    def default(self, o: Any) -> Any:
        # Check if the object is an SQLAlchemy declarative instance
        if isinstance(o.__class__, DeclarativeMeta):
            fields = {}
            # Get all fields of the class
            for field in SQLAIntrospector(o).column_attrs:
                data = self._process_column(o, field)
                fields[field] = data
            return fields

        # Or if it is an SQLAlchemy table instance
        elif isinstance(o, Table):
            fields = {}
            # Get all columns of the table
            for field in [c.key for c in SQLAIntrospector(o).columns]:
                data = self._process_column(o, field)
                fields[field] = data
            return fields

        # If it is not an SQLAlchemy object, use the default encoder
        return super(SqlaDeclarativeEncoder, self).default(o)

    def _process_column(self, o: Any, column_name: str):
        # Obtain the value of the column and check for custom encoding
        data = self.dump(
            getattr(o, column_name)
        )
        try:  # Try JSON encoding the field
            json.dumps(data)
        except TypeError as exc:  # if it fails resort to failure hook
            data = self.handle_failure(exc, data)
        return data

    def dump(self, v: Any) -> Any:
        """ Hook for custom object deserialization """
        if v.__class__ in self.datetypes:
            return self.serialize_datetypes_to_iso(v)
        return v

    # noinspection PyMethodMayBeStatic
    def serialize_datetypes_to_iso(self, v: datetypes) -> str:
        """ Converts datetime formats to ISO8601 compliant strings """
        if isinstance(v, datetime.timedelta):
            return (datetime.datetime.min + v).time().isoformat()
        return v.isoformat()

    def handle_failure(self, exc: Exception, value: Any):
        """ Can be overridden to provide handling for custom fields """
        # Simply re-raise the exception by default
        raise exc


__all__ = ["SqlaDeclarativeEncoder"]



import os
import pathlib
import shutil
import re
from collections import OrderedDict
from functools import wraps
from typing import Union, Optional, Iterable
from urllib.parse import urlparse
from uuid import UUID


def get_application_root() -> str:
    """ Get the root path of the running application """
    return os.environ.get("PYTHONPATH").split(os.pathsep)[0]


def is_dunder(name: str) -> bool:
    """ Check whether a string is a dunder function name """
    if (name[:2] and name[-2:]) in ("__",):
        return True
    return False


def all_items_equal(iterable: Iterable) -> bool:
    """ Check whether all items in a non-hashable iterable are equal """
    if isinstance(iterable, (list, tuple)):
        return True if len(set(iterable)) == 1 else False
    raise TypeError("Only list and tuples may be processed")


def all_nested_zipped_equal(iterable: Iterable[Iterable]) -> bool:
    """
    Check whether all items at the same indexes
    in a nested iterable are equal.

    Example\n
    mylist = [[1, 2], [1, 2], [1, 2]]\n
    Would check\n
    mylist[0][0] == mylist[1][0] == mylist[2][0]
    """
    return all(
        [all_items_equal(subiter) for subiter in zip(*iterable)]
    )


def uuid4_is_valid(uuid: str) -> bool:
    """ Check whether a given string is a valid UUIDv4 """
    try:
        val = UUID(uuid, version=4)
    except ValueError:
        return False
    return val.hex == uuid


def exclude_keys(dictionary: dict, keys: Iterable) -> dict:
    """ Return a new dict - the specified keys """
    return {
        k: v for k, v in dictionary.items() if k not in keys
    }


def all_keys_present(dictionary: dict, keys: Iterable) -> bool:
    """ Check whether all specified keys are present in the dict """
    return all(k in dictionary.keys() for k in keys)


def all_items_present(list_: Union[list, tuple], values: Iterable) -> bool:
    """ Check whether all provided values are present in the list """
    return all(k in list_ for k in values)


def url_is_http(url: str) -> bool:
    """ Check whether a URL is valid HTTP """
    if url.startswith("http") or url.startswith("https"):
        try:
            result = urlparse(url)
            return all([result.scheme, result.netloc, result.path])
        except (AttributeError, TypeError):
            return False
    return False


def memoize(fn):
    """
    Caches the last passed parameter,
    until a new one is provided.\n
    Only works for function with a single positional param.

    Probably not thread-safe, so use with caution.
    """
    # use a dict for mutability
    cache = OrderedDict()

    @wraps(fn)
    def mem_f(arg=None):
        if arg:
            cache[0] = arg
        try:
            res = fn(cache[0])
        except KeyError:
            raise TypeError(
                f"Function {fn.__name__} missing positional argument"
            ) from None
        else:
            return res

    return mem_f


class AttrDict(dict):
    """ A dictionary whose keys can be accessed like attributes """
    __slots__ = []

    def __getattr__(self, item):
        try:
            return super(AttrDict, self).__getitem__(item)
        except KeyError:
            # we need to do this so attribute access
            # does not throw a key error
            raise AttributeError from None


class _TfvarsParser:
    """ A parser class for *.tfvars files """
    def __init__(self, path: pathlib.Path) -> None:
        self.path = path

    def read(self) -> dict:
        with open(self.path, "r") as fin:
            c = [
                line for line in fin.readlines() if not self._ignore_line(line) and line
            ]

        attrs = {
            self._make_name_safe(l[0]): self._guess_type(l[1]) for l in [
                [e.strip() for e in line.split("=")] for line in c
            ]
        }
        return attrs

    # noinspection PyMethodMayBeStatic
    def _ignore_line(self, line: str) -> bool:
        # Case for comments
        if line.strip().startswith(("/", "*", "#")):
            return True

        # Cannot be a key value block if there is no '='
        if "=" not in line:
            return True

        return False

    # noinspection PyMethodMayBeStatic
    def _make_name_safe(self, key: str) -> str:
        # If it is a key value block, any - must be converted to _
        return key.replace("-", "_").lower()

    # noinspection PyMethodMayBeStatic
    def _guess_type(self, val: str) -> Union[str, int, float]:
        if val.isdigit():
            return int(val)
        else:
            try:
                return float(val)
            except ValueError:
                pass
        if any(v in val for v in ["'", '"']):
            return val[1:-1]
        return val


class Tfvars:
    """ Wrapper for a parsed *.tfvars file """
    def __init__(self, /, path: Union[str, pathlib.Path]) -> None:
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        if not path.exists():
            raise FileNotFoundError(f"The directory {path} does not exist")
        self.path = path
        self.vars = AttrDict(**_TfvarsParser(path).read())


# this is necessary because we can only subclass the concrete implementation
class Path(type(pathlib.Path())):
    def rmdir(self, recursive: Optional[bool] = True) -> None:
        if recursive:
            shutil.rmtree(self)
        else:
            super(Path, self).rmdir()


__all__ = [
    "is_dunder", "get_application_root", "uuid4_is_valid",
    "all_items_equal", "all_nested_zipped_equal", "all_items_present",
    "exclude_keys", "all_keys_present", "url_is_http",
    "AttrDict", "Tfvars", "Path", "memoize"
]



import weakref
from re import sub
from typing import Any

from sqlalchemy import MetaData, Integer, Column, Sequence
from sqlalchemy.ext.declarative import (
    declarative_base,
    declared_attr,
    DeclarativeMeta
)


def gen_tablenames(name: str) -> str:
    """ Converts CamelCase class names to snake_case table names """
    return sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


class _declared_Base:
    @declared_attr
    def __tablename__(self) -> str:
        """ Automatically sets the name for created tables """
        return gen_tablenames(self.__name__)

    pk = Column(
        Integer(),
        Sequence("%(table_name)s_id_seq"),
        primary_key=True
    )


Base = declarative_base(
    cls=_declared_Base,
    metadata=MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",  # index
            "uq": "uq_%(table_name)s_%(column_0_name)s",  # unique constraint
            "ck": "ck_%(table_name)s_%(constraint_name)s",  # check constraint
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",  # foreign key
            "pk": "pk_%(table_name)s",  # primary key
        }
    ),
)


class Registry:
    def __init__(self, base) -> None:
        self._base = weakref.ref(base)
        self.tables = base._decl_class_registry  # noqa access to protected

    def table_class_from_tablename(self, tablename: str) -> DeclarativeMeta:
        base = self._resolve_weakref(self._base)
        assert issubclass(base, Base), \
            "This method attempts to reverse the automatic " \
            "naming done by this packages custom " \
            "declarative Base." \
            "When other baseclasses are used this process " \
            "does not work."
        refs = {
            gen_tablenames(c): c for c in self.tables.keys()
        }
        classname = refs[tablename]
        return self._resolve_weakref(self.tables[classname])

    # noinspection PyMethodMayBeStatic
    def _resolve_weakref(self, ref: weakref.ref) -> Any:
        return ref()

    def __getitem__(self, table_classname: str) -> DeclarativeMeta:
        table_ref = self.tables[table_classname]
        return self._resolve_weakref(table_ref)


__all__ = ["gen_tablenames", "Base", "Registry"]



from .conn import Database
from .inspect import SQLAIntrospector
from .base import Base, Registry
from .encode import SqlaDeclarativeEncoder

__all__ = [
    "Database",
    "Base", "Registry",
    "SQLAIntrospector",
    "SqlaDeclarativeEncoder"
]



from contextlib import contextmanager
from logging import getLogger, Logger
from typing import Iterator, Tuple, ClassVar

from sqlalchemy import create_engine
from sqlalchemy.orm import Session as sqlaSession, sessionmaker
from sqlalchemy.pool import StaticPool

from .base import Base


class Database:
    __slots__ = ["engine", "_Session"]
    _Session: sessionmaker
    logger: ClassVar[Logger] = getLogger()

    def __init__(self, /,
                 conn_args: dict = None, *,
                 dialect: str = None,
                 host: str = None,
                 username: str = None,
                 password: str = None,
                 port: int = None,
                 database_name: str = None,
                 extra_engine_args: dict = None) -> None:
        kwargs = locals()
        kwargs.pop("conn_args")
        kwargs = {k: v for k, v in kwargs.items() if v}
        assert kwargs or conn_args, \
            "Specify either conn_args or one of the kwargs"
        config = conn_args or kwargs

        uri, engine_args = self._get_connection_conf(config)
        self.engine = create_engine(uri, **engine_args)
        self._Session = sessionmaker(
            autocommit=False,
            autoflush=False,
            expire_on_commit=False,
            bind=self.engine,
        )

    # noinspection PyMethodMayBeStatic
    def _get_connection_conf(self, config) -> Tuple[str, dict]:
        """
        Automatically performs optimal parametrization of the SQLAlchemy engine.\n
        If this functionality is not desired, a manual override of this function,
        can be performed using the 'override_engine' function.

        :return: Tuple containing the SQLAlchemy database uri and engine parameters
        """
        # Make sure that at least the dialect option exists
        assert config.get("dialect", False), "Database dialect selection is mandatory"
        # If host is empty and the dialect is 'sqlite', we know the user wants :memory:
        dialect, host = config.get("dialect"), config.get("host", None)
        engine_args: dict = {"pool_pre_ping": True}
        if dialect == "sqlite":
            engine_args.update({"connect_args": {"check_same_thread": False}})
            if not host:
                engine_args.update({"poolclass": StaticPool})
            uri: str = f"{dialect}:///{host if host else ':memory:'}"
        # This is the option for all other database types, such as Postgres or MySQL
        else:
            username, password, port, database = tuple(
                config.get(attr, None) for attr in (
                    "username", "password", "port", "database"
                )
            )
            if not username and not port:
                raise ValueError("Credentials and / or port missing")
            uri: str = f"{dialect}:///{username}:{password}@{host}:{port}/{database}"
        return uri, engine_args

    def override_engine(self, uri, **kwargs):
        """
        Provides an option to manually override the auto-specced engine.\n

        :param uri: Database URI
        :param kwargs: Normal kwargs as provided to the create_engine factory
        """
        self.engine = create_engine(uri, **kwargs)

    @contextmanager
    def Session(self) -> Iterator[sqlaSession]:
        """ Provides access to a scoped ORM Session """
        session: sqlaSession = self._Session()  # noqa
        try:
            yield session
            session.commit()
        except Exception as exc:  # noqa E722
            # In case there is an error during execution we rollback,
            # so no corrupted data reaches the database or fills the cache
            session.rollback()
            raise exc
        finally:
            # Always close the session after operations
            session.close()

    def create(self, base_override=None) -> None:
        """ Creates all tables in the current Base """
        base = base_override or Base
        self.logger.info(
            f"Constructing Database: {str(self.engine.url)}"
        )
        for table in base.metadata.sorted_tables:
            self.logger.info(f"     {table.name}")
        base.metadata.create_all(bind=self.engine)

    def destroy(self, base_override=None) -> None:
        """ Destroys all tables in the current Base """
        base = base_override or Base
        base.metadata.drop_all(bind=self.engine)


__all__ = ["Database"]



import requests
import requests.adapters
import google.auth.exceptions
from google.oauth2 import service_account
from google.oauth2._client import id_token_jwt_grant  # noqa protected
from google.auth.transport.requests import Request


def construct_service_authentication_request() -> Request:
    auth_request_session = requests.Session()
    retry_adapter = requests.adapters.HTTPAdapter(max_retries=3)
    auth_request_session.mount("https://", retry_adapter)
    auth_request = Request(auth_request_session)
    return auth_request


def force_obtain_id_token(credentials: service_account.IDTokenCredentials) -> str:
    """ Manually forces the equivalent of credentials.refresh() """
    assertion = credentials._make_authorization_grant_assertion()  # noqa protected
    request = construct_service_authentication_request()
    try:
        access_token, *_ = id_token_jwt_grant(
            request,
            credentials._token_uri,  # noqa protected
            assertion  # noqa expected type
        )
    except google.auth.exceptions.RefreshError as exc:
        raise Exception(
            "Error when requesting the token, "
            "you may have provided an empty target_audience parameter,"
            " for the Credentials object."
        ) from exc
    return access_token



from abc import ABC, abstractmethod
from inspect import isclass
from typing import Union

from sqlalchemy import Table
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.orm.util import class_mapper

from ..utils import is_dunder


def is_mapped_class(cls) -> bool:
    if not isclass(cls):
        return False
    try:
        class_mapper(cls)
    except UnmappedClassError:
        return False
    else:
        return True


class _IntrospectorInterface(ABC):
    tablename: str
    columns: dict
    __table__: Table

    def __init__(self, obj) -> None:
        self.schema = obj

    @property
    @abstractmethod
    def classname(self):
        pass

    @property
    def column_attrs(self):
        remove_keys = (
            "metadata",  # ref to DeclarativeMeta
            "_decl_class_registry",  # declarative registry
            "_sa_class_manager"  # class state manager
        )
        attrs = [
            attr for attr in dir(
                self.schema
            ) if not is_dunder(
                attr
            ) and not callable(
                getattr(self.schema, attr)
            ) and attr not in remove_keys
        ]
        return attrs

    def primary_keys(self, retrieve_constraint: bool = False) -> list:
        constraints = inspect(self.__table__).primary_key
        if retrieve_constraint:
            return constraints
        return [c.key for c in constraints.columns]

    def unique_keys(self) -> list:
        return [
            c.name for c in self.columns if any(
                [c.primary_key, c.unique]
            )
        ]

    def foreign_keys(self, columns_only: bool = False) -> Union[list, dict]:
        foreign_keys = [c for c in self.columns if c.foreign_keys]
        fk_names = [key.name for key in foreign_keys]
        if columns_only:
            return fk_names
        fk_targets = [
            list(key.foreign_keys)[0].target_fullname for key in foreign_keys
        ]
        return {
            name: target for name, target in zip(fk_names, fk_targets)
        }


class _declBaseIntrospector(_IntrospectorInterface):
    """
    Inspect Declarative Base class instances.
    These are objects you may get back when querying,
    using a Session or when constructing an instance yourself
    """
    def __init__(self, obj) -> None:
        super().__init__(obj)
        self.tablename = obj.__tablename__
        self.columns = obj.__table__.c
        self.__table__ = obj.__table__

    @property
    def classname(self) -> str:
        return self.schema.__class__.__name__

    @property
    def column_attrs(self):
        attrs = super().column_attrs
        remove_keys = (
            "_sa_instance_state",  # instance session state manager
        )
        return [attr for attr in attrs if attr not in remove_keys]


class _declClassIntrospector(_IntrospectorInterface):
    def __init__(self, obj) -> None:
        super().__init__(obj)
        self.tablename = obj.__tablename__
        self.columns = obj.__table__.c
        self.__table__ = obj.__table__

    @property
    def classname(self) -> str:
        return self.schema.__name__


class _tblIntrospector(_IntrospectorInterface):
    def __init__(self, obj: Table) -> None:
        super().__init__(obj)
        self.tablename = obj.name
        self.columns = obj.c  # noqa false-positive write only property
        self.__table__ = obj

    @property
    def classname(self):
        raise NotImplementedError(
            f"Class {self.__class__.__name__} does not support "
            f"declarative base features"
        )

    @property
    def column_attrs(self):
        raise NotImplementedError(
            f"Autogenerated table instances do not support declarative "
            f"class properties or hybrid properties"
        )


def SQLAIntrospector(o):
    """ Factory returning a matching introspector class """
    err_msg = "Class of type {0} is not supported for introspection"
    if isinstance(o, Table):
        return _tblIntrospector(o)

    if isclass(o):
        if is_mapped_class(o):
            return _declClassIntrospector(o)
        raise TypeError(err_msg.format(type(o)))
    else:
        if is_mapped_class(o.__class__):
            return _declBaseIntrospector(o)
        raise TypeError(err_msg.format(type(o.__class__)))


__all__ = ["SQLAIntrospector", "is_mapped_class"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drizm_commons', 'drizm_commons.google', 'drizm_commons.sqla']

package_data = \
{'': ['*']}

extras_require = \
{'google': ['google-auth==1.23.0', 'requests>=2.25.0,<3.0.0'],
 'sqla': ['sqlalchemy==1.3.20']}

setup_kwargs = {
    'name': 'drizm-commons',
    'version': '0.4.0',
    'description': 'Python3 commons for the Drizm organization',
    'long_description': '# Python Commons\n[![PyPI version](https://badge.fury.io/py/drizm-commons.svg)](https://badge.fury.io/py/drizm-commons)  \n\nThis package includes shared code used by\nthe Drizm organizations development team.  \n\nIt is not intended for public usage but you\nmay still download, redistribute or \nmodify it to your liking.\n\n## Usage\n\nBasic Install (utils only):  \n>pip install drizm-commons\n\n\nFull install (SQLAlchemy features available):  \n>pip install drizm-commons[sqla]\n\nImport like so:  \n*import drizm_commons*\n\n## Documentation\n\n### Google Cloud Tools\n\n````python\nfrom drizm_commons.google import force_obtain_id_token\nfrom google.oauth2 import service_account\n\n\nauth = service_account.IDTokenCredentials.from_service_account_file(\n    "/path/to/svc.json",\n    target_audience="https://example.com/"\n)\ntoken = force_obtain_id_token(auth)\n# returns something like \'ey....\', only access token\n````\n\n### Utilities\n\n**Convinience Functions:**  \n````python\nfrom drizm_commons.utils import *\n\n\n# Check whether function name is dunder\nis_dunder("__name__")  # True\n\n# Check if a given string is a valid UUIDv4\nuuid4_is_valid("myvalue")  # False\n\n# Check if a URL is valid and the contents URL-Safe\nurl_is_http("https://myapp.com/")  # True\n\n# Get the current applications root path\nPath(get_application_root())\n````\n\n**Path with extra features:**\n````python\nfrom drizm_commons.utils import Path\n\n# Recursively delete a folder\npath = Path(__file__).parent\npath.rmdir(recursive=True)\n````\n\n**Cache last passed parameter:**\n````python\nfrom drizm_commons.utils import memoize\n\n\n@memoize\ndef foo(a):\n    return a\n\n\nfoo(3)  # 3\nfoo()  # 3\n````\n\n### Introspection\n\n````python\nfrom drizm_commons.inspect import SQLAIntrospector\n\n\ntable = SQLAIntrospector(my_table_instance)\n\n""" Attributes """\ntable.tablename  # get the name of the table\ntable.classname  # get the classname of the declarative instance\ntable.columns  # get all SQLA fields of the class\ntable.column_attrs  # get all SQLA fields + property and hybrid_property of the class\n````\n\n## Changelog\n\n### 0.1.1\n\n- Added SQLAlchemy JSON Encoder\n- Fixed bugs related to the Introspection\nAPI\n- Added table registry\n- Added additional utilities\n\n### 0.1.2\n\n- Added get_root_path and recursive delete\nPath utilities\n- Fixed various bugs\n\n### 0.2.0\n\n- Added full test suite\n- Added testing tools\n- Revamped introspection API\n- Provided additional overrides for the\nSQL connection adapter\n\n### 0.2.1\n\n- Added support for datetime JSON\nencoding\n\n### 0.2.2\n\n- Improved in-code documentation\n- Integrated additional utils from\ndrizm-django-commons\n\n### 0.3.0\n\n- Added introspection capabilities \nfor property and SQLAlchemy\'s\nhybrid_property\n- SQLAEncoder now respects property\nand hybrid_property on SQLA declarative\ninstances\n- Additional customizability hooks\nfor custom fields or data handling\n- Support for JSON-Encoding table\ninstances\n- Added SQLA as optional dependency\n- Added additional testing utilities\n\n### 0.3.1\n\n- Improved code documentation\n- Added docs\n- Added memoize function decorator\nto cache last previously passed\nfunction parameter\n\n### 0.3.2\n\n- Fixed issue with introspection API\npicking up validation methods\n\n### 0.3.3\n\n- Added additional tests and bugfixes\n\n### 0.3.4\n\n- Added support for comments and\nspecial character parsing to Tfvars\n\n### 0.4.0\n\n- Added method to force obtain\nGoogleCloudPlatform Id-Tokens\n',
    'author': 'ThaRising',
    'author_email': 'kochbe.ber@gmail.com',
    'maintainer': 'Dominik Lewandowski',
    'maintainer_email': 'dominik.lewandow@gmail.com',
    'url': 'https://github.com/drizm-team/python-commons',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
